/**
 * Copyright (C) 2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.server.session;

import java.util.ArrayList;
import java.util.List;

import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerFactory;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.session.RunningSync;
import com.funambol.framework.server.session.RunningSyncException;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.server.store.RunningSyncPersistentStore;
/**
 * The manager of the running syncs.
 * @version $Id: RunningSyncManager.java,v 1.3 2007/04/06 17:46:52 nichele Exp $
 */
public class RunningSyncManager {

    // --------------------------------------------------------------- Constants

    public static final String SYSTEM_PROPERTY_NODE_ID = "funambol.ds.node.id";

    private static final String LOGGER_NAME                     = "funambol.server.runningsync";
    private static final String DEFAULT_NODE_ID                 = "DEFAULT_NODE";
    private static final int    DEFAULT_MAX_SYNC_AGE_IN_MINUTES = 480;  // 8 hours

    // ------------------------------------------------------------ Private data

    /**
     * The used logger
     */
    private FunambolLogger logger = null;

    // -------------------------------------------------------------- Properties

    /**
     * The node id assigned to the ds server
     */
    private String nodeId = null;

    public String getNodeId() {
        return nodeId;
    }

    public void setNodeId(String nodeId) {
        this.nodeId = nodeId;
    }

    /**
     * The max age of a sync in minutes and in millis
     */
    private int maxSyncAgeInMinutes = DEFAULT_MAX_SYNC_AGE_IN_MINUTES;
    private long maxSyncAgeInMillis = maxSyncAgeInMinutes * 60 * 1000;

    public int getMaxSyncAgeInMinutes() {
        return maxSyncAgeInMinutes;
    }

    public void setMaxSyncAgeInMinutes(int maxSyncAgeInMinutes) {
        this.maxSyncAgeInMinutes = maxSyncAgeInMinutes;
        this.maxSyncAgeInMillis  = maxSyncAgeInMinutes * 60 * 1000;
    }

    /**
     * The persistent store used to handle the running sync
     */
    private RunningSyncPersistentStore persistentStore = null;

    public RunningSyncPersistentStore getPersistentStore() {
        return persistentStore;
    }

    public void setPersistentStore(RunningSyncPersistentStore persistentStore) {
        this.persistentStore = persistentStore;
    }

    // ------------------------------------------------------------ Constructors

    /** Creates a new instance of RunningSyncManager */
    public RunningSyncManager() {
        logger = FunambolLoggerFactory.getLogger(LOGGER_NAME);
        nodeId  = System.getProperty(SYSTEM_PROPERTY_NODE_ID);
        if (nodeId == null) {
            nodeId = DEFAULT_NODE_ID;
        }
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Removes all running syncs with the node id assigned to this instance of
     * RunningSyncManager
     * @throws com.funambol.framework.server.session.RunningSyncException if an error occurs
     */
    public void removeAllRunningSyncByNode() throws RunningSyncException {
        if (logger.isTraceEnabled()) {
            logger.trace("Deleting all running sync entry with node: " + nodeId);
        }
        try {
            persistentStore.deleteByNode(nodeId);
        } catch (PersistentStoreException ex) {
            throw new RunningSyncException("Error deleting the running sync " +
                                           "for the node '" + nodeId + "'", ex);
        }
        if (logger.isTraceEnabled()) {
            logger.trace("All running sync entry with node '" + nodeId + "' deleted");
        }
    }

    /**
     * Removes all running syncs with the given username
     * @param username the username
     * @throws com.funambol.framework.server.session.RunningSyncException if an error occurs
     * @return all the running syncs of the given user
     */
    public RunningSync getRunningSyncByUsername(String username)
    throws RunningSyncException {

        RunningSync[] syncs = null;
        try {
            syncs = (RunningSync[])persistentStore.readByUsername(username);
            if (syncs != null && syncs.length > 0) {
                //
                // Deleting the old sync (dead sync)
                //
                syncs = removingDeadSync(syncs);
                if (syncs != null && syncs.length > 0) {
                    return syncs[0];
                }
            }
        } catch (PersistentStoreException ex) {
            throw new RunningSyncException("Error checking if there is already a running sync for the user '" +
                                           username + "'", ex);
        }
        return null;
    }

    /**
     * Inserts a new running sync with the given properties.
     * <br/>
     * Note the method calls persistentStore.insert(sync) so, if there is already
     * a sync with the same username, a RunningSyncException is thrown.
     *
     * @param sessionId the session id
     * @param principal the principal
     * @param startSync the time
     * @throws RunningSyncException if an error occurs
     * @return true is there is not an already running sync for the same user, false otherwise
     */
    public boolean startSync(String sessionId, Sync4jPrincipal principal, long startSync)
    throws RunningSyncException {

        RunningSync sync = new RunningSync(sessionId,
                                           nodeId,
                                           principal.getId(),
                                           principal.getUsername(),
                                           principal.getDeviceId(),
                                           startSync);
        try {
            persistentStore.insert(sync);
        } catch (PersistentStoreException ex) {
            throw new RunningSyncException("Error storing the running sync", ex);
        }

        return true;
    }

    /**
     * Stores a running sync with the given properties.
     * <br/>
     * Note the method calls persistentStore.store(sync) that, first of all, tries
     * with an update, and then, it performs an insert
     *
     * @param sessionId the session id
     * @param principal the principal
     * @param startSync the time
     * @throws RunningSyncException if an error occurs
     * @return true is there is not an already running sync for the same user, false otherwise
     */
    public boolean storeSync(String sessionId, Sync4jPrincipal principal, long startSync)
    throws RunningSyncException {

        RunningSync sync = new RunningSync(sessionId,
                                           nodeId,
                                           principal.getId(),
                                           principal.getUsername(),
                                           principal.getDeviceId(),
                                           startSync);
        try {
            persistentStore.store(sync);
        } catch (PersistentStoreException ex) {
            throw new RunningSyncException("Error storing the running sync", ex);
        }

        return true;
    }

    /**
     * Deletes from the running syncs list the entry with the given principal
     * @param principal the principal
     * @throws RunningSyncException if an error occurs
     */
    public void endSync(String sessionId)
    throws RunningSyncException {
        try {
            persistentStore.deleteBySessionId(sessionId);
        } catch (PersistentStoreException ex) {
            throw new RunningSyncException("Error deleting the running sync " +
                                           "for the sessionId '" + sessionId + "'", ex);
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Removes the dead syncs from the given array.
     * A dead sync is a sync that is old more than maxSyncAgeInMinutes.
     * @param syncs an array of RunningSync
     * @return an array of RunningSync with the running sync still alive
     */
    private RunningSync[] removingDeadSync(RunningSync[] syncs) {
        if (syncs == null || syncs.length == 0) {
            return syncs;
        }
        List<RunningSync> aliveSyncs = new ArrayList();
        long currentTime = System.currentTimeMillis();
        for (RunningSync sync : syncs) {
            long startSync = sync.getStartSync();
            if ( (currentTime - startSync) > maxSyncAgeInMillis ) {
                String sessionId = null;
                try {
                    //
                    // Dead sync
                    //
                    sessionId = sync.getSessionId();
                    if (logger.isTraceEnabled()) {
                        logger.trace("Deleting dead sync with session '" + sessionId + "'");
                    }
                    persistentStore.deleteBySessionId(sessionId);
                } catch (PersistentStoreException ex) {
                    logger.error("Error deleting the dead sync with session '" +
                                 sessionId + "'", ex);
                }
            } else {
                aliveSyncs.add(sync);
            }
        }
        return (RunningSync[])aliveSyncs.toArray(new RunningSync[aliveSyncs.size()]);
    }
}
