/**
 * Copyright (C) 2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.store;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.funambol.framework.filter.Clause;
import com.funambol.framework.filter.WhereClause;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.session.RunningSync;
import com.funambol.framework.server.store.*;
import com.funambol.framework.tools.DBTools;
import com.funambol.framework.tools.beans.LazyInitBean;

/**
 * This is the store for running sync information.
 * Currently it persistes the following classes:
 * <ul>
 * <li>com.funambol.framework.server.session.RunningSync</li>
 * </ul>
 *
 * @version $Id: RunningSyncPersistentStore.java,v 1.1 2007/04/06 16:39:37 nichele Exp $
 */
public class RunningSyncPersistentStore
extends BasePersistentStore
implements PersistentStore, java.io.Serializable, LazyInitBean {

    // --------------------------------------------------------------- Constants

    public static final int SQL_INSERT                 = 0;
    public static final int SQL_UPDATE                 = 1;
    public static final int SQL_SELECT_ALL             = 2;
    public static final int SQL_SELECT                 = 3;
    public static final int SQL_COUNT                  = 4;
    public static final int SQL_DELETE_USERNAME        = 5;
    public static final int SQL_DELETE_PRINCIPAL       = 6;
    public static final int SQL_DELETE_NODE            = 7;
    public static final int SQL_DELETE_SESSIONID       = 8;

    // -------------------------------------------------------------- Properties

    protected String[] sql = null;

    public void setSql(String[] sql) {
        this.sql = sql;
    }

    public String[] getSql() {
        return this.sql;
    }

    // ------------------------------------------------------------ Private data

    // ------------------------------------------------------------ Constructors

    // ---------------------------------------------------------- Public methods
    public void init() {
        //
        // Nothing to do
        //
    }

    public void configure(Map properties) throws ConfigPersistentStoreException {
        //
        // Nothing to do
        //
    }

    /**
     * Delete the running sync from the data store.
     * @param o The object that must be a RunningSync for this method
     * @return treu if the given object is a <CODE>RunningSync</CODE>, false otherwise
     * @throws PersistentStoreException if an error occurs
     */
    public boolean delete(Object o) throws PersistentStoreException {
        if (o instanceof RunningSync) {
            RunningSync runningSync = (RunningSync) o;

            deleteByUsername(runningSync.getUsername());

            return true;
        }
        return false;
    }

    /**
     * Deletes the running syncs with the given principal
     * @param principal the principal
     * @return the numbers of the deleted row (by design, can be 0 or 1 because in the database
     * the username is key)
     * @throws PersistentStoreException if an error occurs
     */
    public int deleteByPrincipal(Sync4jPrincipal principal) throws PersistentStoreException {
        Connection        conn = null;
        PreparedStatement stmt = null;

        int numOfDeletion = 0;

        try {
            conn = dataSource.getConnection();

            stmt = conn.prepareStatement(sql[SQL_DELETE_PRINCIPAL]);
            stmt.setLong(1, principal.getId());
            numOfDeletion = stmt.executeUpdate();

        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the running sync " +
                                               "for the principal '" + principal + "'", e);
        } finally {
            DBTools.close(conn, stmt, null);
        }
        return numOfDeletion;
    }

    /**
     * Deletes the running syncs with the given session id.
     * @param sessionId the session id
     * @throws PersistentStoreException if an error occurs
     * @return the numbers of the deleted rows (it should be 0 or 1 because the session id should
     * be unique)
     */
    public int deleteBySessionId(String sessionId) throws PersistentStoreException {
        Connection        conn = null;
        PreparedStatement stmt = null;

        int numOfDeletion = 0;

        try {
            conn = dataSource.getConnection();

            stmt = conn.prepareStatement(sql[SQL_DELETE_SESSIONID]);
            stmt.setString(1, sessionId);
            numOfDeletion = stmt.executeUpdate();

        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the running sync " +
                                               "for the session '" + sessionId + "'", e);
        } finally {
            DBTools.close(conn, stmt, null);
        }

        return numOfDeletion;
    }

    /**
     * Deletes the running syncs with the given username.
     * @param username the username
     * @throws PersistentStoreException if an error occurs
     * @return the numbers of the deleted rows (it should be 0 or 1 because the username is the
     * primary key)
     */
    public int deleteByUsername(String username) throws PersistentStoreException {
        Connection        conn = null;
        PreparedStatement stmt = null;

        int numOfDeletion = 0;

        try {
            conn = dataSource.getConnection();

            stmt = conn.prepareStatement(sql[SQL_DELETE_USERNAME]);
            stmt.setString(1, username);
            numOfDeletion = stmt.executeUpdate();

        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the running sync " +
                                               "for the user '" + username + "'", e);
        } finally {
            DBTools.close(conn, stmt, null);
        }

        return numOfDeletion;
    }

    /**
     * Deletes the running syncs with the given node
     * @param node the node id
     * @throws PersistentStoreException if an error occurs
     * @return the numbers of the deleted rows
     */
    public int deleteByNode(String node) throws PersistentStoreException {
        Connection        conn = null;
        PreparedStatement stmt = null;

        int numOfDeletion = 0;

        try {
            conn = dataSource.getConnection();

            stmt = conn.prepareStatement(sql[SQL_DELETE_NODE]);
            stmt.setString(1, node);
            numOfDeletion = stmt.executeUpdate();

        } catch (SQLException e) {
            throw new PersistentStoreException("Error deleting the running sync " +
                                               "for the node '" + node + "'", e);
        } finally {
            DBTools.close(conn, stmt, null);
        }

        return numOfDeletion;
    }

    /**
     * Stores the given object if it is a RunningSync
     * @param o the object to save
     * @return true if the object is saved, false otherwise.
     * @throws PersistentStoreException if an error occurs
     */
    public boolean store(Object o) throws PersistentStoreException {
        if (o instanceof RunningSync) {
            RunningSync runningSync = (RunningSync) o;

            Connection        conn = null;
            PreparedStatement stmt = null;

            int n = 0;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sql[SQL_UPDATE]);
                stmt.setString(1, runningSync.getSessionId());
                stmt.setString(2, runningSync.getNode());
                stmt.setLong  (3, runningSync.getPrincipal());
                stmt.setString(4, runningSync.getDevice());
                stmt.setLong  (5, runningSync.getStartSync());

                stmt.setString(6, runningSync.getUsername());

                n = stmt.executeUpdate();
                stmt.close(); stmt = null;

                if (n == 0) {
                    // The first time!!!
                    stmt = conn.prepareStatement(sql[SQL_INSERT]);
                    stmt.setString(1, runningSync.getUsername());
                    stmt.setString(2, runningSync.getSessionId());
                    stmt.setString(3, runningSync.getNode());
                    stmt.setLong  (4, runningSync.getPrincipal());
                    stmt.setString(5, runningSync.getDevice());
                    stmt.setLong  (6, runningSync.getStartSync());
                    stmt.executeUpdate();
                }
                return true;
            } catch (Exception e) {
                throw new PersistentStoreException("Error storing running sync " + runningSync, e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
        }
        return false;
    }

    /**
     * Read the running sync from the data store.
     * @param o The object that must be a RunningSync for this method
     * @return true if the given object is a <CODE>RunningSync</CODE>, false otherwise
     * @throws PersistentStoreException if an error occurs
     */
    public boolean read(Object o) throws PersistentStoreException {
        if (o instanceof RunningSync) {
            RunningSync runningSync = (RunningSync) o;

            Connection        conn = null;
            PreparedStatement stmt = null;
            ResultSet         rs   = null;

            try {
                conn = dataSource.getConnection();
                String sessionId = runningSync.getSessionId();

                stmt = conn.prepareStatement(sql[SQL_SELECT]);
                stmt.setString(1, runningSync.getUsername());
                rs = stmt.executeQuery();

                if (rs.next() == false) {
                    throw new NotFoundException("Running sync with session id '" +
                                                sessionId + "' not found");
                }

                runningSync.setUsername (rs.getString(1));
                runningSync.setSessionId(rs.getString(2));
                runningSync.setNode     (rs.getString(3));
                runningSync.setPrincipal(rs.getLong(4));
                runningSync.setDevice   (rs.getString(5));
                runningSync.setStartSync(rs.getLong(6));

            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading running sync " + runningSync, e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
            return true;
        }
        return false;
    }

    /**
     * Inserts the given RunningSync
     * @param runningSync the runningSync to save
     * @throws PersistentStoreException if an error occurs
     */
    public void insert(RunningSync runningSync) throws PersistentStoreException {

        Connection        conn = null;
        PreparedStatement stmt = null;
        ResultSet         rs   = null;
        int n = 0;

        try {
            conn = dataSource.getConnection();

            stmt = conn.prepareStatement(sql[SQL_INSERT]);
            stmt.setString(1, runningSync.getUsername());
            stmt.setString(2, runningSync.getSessionId());
            stmt.setString(3, runningSync.getNode());
            stmt.setLong  (4, runningSync.getPrincipal());
            stmt.setString(5, runningSync.getDevice());
            stmt.setLong  (6, runningSync.getStartSync());
            stmt.executeUpdate();

        } catch (Exception e) {
            throw new PersistentStoreException("Error inserting running sync " + runningSync, e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }

    /**
     * Returns the RunningSync of the given user
     * @param username the username
     * @return an array containing the RunningSync with the given username (by design, the array
     * is empty or with just an entry because the username is primary key)
     * @throws PersistentStoreException if an error occurs
     */
    public RunningSync[] readByUsername(String username) throws PersistentStoreException {
        //
        // Checking if there is an already running sync for this user
        //
        WhereClause searchByUser = new WhereClause("username",
                                                   new String[] {username},
                                                   WhereClause.OPT_EQ,
                                                   true);

        return read(searchByUser);
    }

    /**
     * Not implemented.
     * @param objClass 
     * @return Always null
     * @throws PersistentStoreException 
     */
    public Object[] read(Class objClass) throws PersistentStoreException {
        return null;
    }

    /**
     * Read all running syncs that satisfy the conditions specified in input.
     * 
     * @return Object[] array of RunningSyncs
     * @param o The object that must be a RunningSync for this method
     * @param clause condition where for select
     * @throws PersistentStoreException if an error occurs
     */
    public Object[] read(Object o, Clause clause) throws PersistentStoreException {
        if (!(o instanceof RunningSync)) {
            return null; // Not me!!
        }
        return read(clause);
    }



    /**
     * Counts the number of running syncs that satisfy the conditions specified in input
     * 
     * @return int number of RunningSyncs that satisfy the conditions
     * @param o The object of type RunningSync.
     * @param clause The additional where clause.
     * @throws PersistentStoreException if an error occurs
     */
    public int count(Object o, Clause clause) throws PersistentStoreException {
        if (!(o instanceof RunningSync)) {
            return -1; // Not me!
        }
        Connection        conn = null;
        PreparedStatement stmt = null;
        ResultSet         rs   = null;
        int n = 0;

        try {
            conn = dataSource.getConnection();

            PreparedWhere where = sqlHelperClause.getPreparedWhere(clause);
            String query = sql[SQL_COUNT];
            if (where.sql.length()>0) {
                StringBuilder sb = new StringBuilder(query);
                sb.append(" where ").append(where.sql);
                query = sb.toString();
            }
            stmt = conn.prepareStatement(query);
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }
            rs = stmt.executeQuery();

            while (rs.next()) {
                n = rs.getInt(1);
            }
            return n;
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading count running sync ", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }

    //---------------------------------------------------------- Private Methods

    /**
     * Read all running syncs that satisfy the conditions specified in input.
     * 
     * @return RunningSync[] array of RunningSyncs that satisfy the conditions
     * @param clause condition where for select
     * @throws PersistentStoreException if an error occurs
     */
    private RunningSync[] read(Clause clause) throws PersistentStoreException {

        Connection        conn = null;
        PreparedStatement stmt = null;
        ResultSet         rs   = null;

        List<RunningSync> ret = new ArrayList();

        try {
            conn = dataSource.getConnection();

            PreparedWhere where = sqlHelperClause.getPreparedWhere(clause);
            String query = sql[SQL_SELECT_ALL];

            if (where.sql.length() > 0) {
                StringBuilder sb = new StringBuilder(query);
                sb.append(" where ").append(where.sql);
                query = sb.toString();
            }
            stmt = conn.prepareStatement(query);

            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }
            rs = stmt.executeQuery();

                while (rs.next()) {
                    RunningSync sync = new RunningSync();
                    sync.setUsername (rs.getString(1));
                    sync.setSessionId(rs.getString(2));
                    sync.setNode     (rs.getString(3));
                    sync.setPrincipal(rs.getLong(4));
                    sync.setDevice   (rs.getString(5));
                    sync.setStartSync(rs.getLong(6));
                    ret.add(sync);
                }

            return ret.toArray(new RunningSync[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading running syncs", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }

    }
}

